@@ -22,6 +22,9 @@ import datatable as dt
 from tqdm import tqdm
 from scipy import sparse
 
+import scanpy as sc
+from .atac import binarize
+
 from typing import Literal
 
 import warnings
@@ -94,13 +97,14 @@ class SURE(nn.Module):
                  input_size: int,
                  codebook_size: int = 200,
                  cell_factor_size: int = 0,
+                 cell_factor_names: list = None,
                  supervised_mode: bool = False,
                  z_dim: int = 10,
                  z_dist: Literal['normal','studentt','laplacian','cauchy','gumbel'] = 'normal',
                  loss_func: Literal['negbinomial','poisson','multinomial','bernoulli'] = 'negbinomial',
                  inverse_dispersion: float = 10.0,
                  use_zeroinflate: bool = True,
-                 hidden_layers: list = [500],
+                 hidden_layers: list = [300],
                  hidden_layer_activation: Literal['relu','softplus','leakyrelu','linear'] = 'relu',
                  nn_dropout: float = 0.1,
                  post_layer_fct: list = ['layernorm'],
@@ -131,6 +135,7 @@ class SURE(nn.Module):
         self.post_layer_fct = post_layer_fct
         self.post_act_fct = post_act_fct
         self.hidden_layer_activation = hidden_layer_activation
+        self.cell_factor_names = cell_factor_names
         
         self.codebook_weights = None
 
@@ -229,8 +234,10 @@ class SURE(nn.Module):
         )
 
         if self.cell_factor_size>0:
-            self.cell_factor_effect = MLP(
-                    [self.cell_factor_size] + self.decoder_hidden_layers + [self.latent_dim],
+            self.cell_factor_effect = nn.ModuleList()
+            for i in np.arange(self.cell_factor_size):
+                self.cell_factor_effect.append(MLP(
+                    [self.z_dim+1] + hidden_sizes + [self.z_dim],
                     activation=activate_fct,
                     output_activation=None,
                     post_layer_fct=post_layer_fct,
@@ -238,6 +245,7 @@ class SURE(nn.Module):
                     allow_broadcast=self.allow_broadcast,
                     use_cuda=self.use_cuda,
                 )
+                )
             
         self.decoder_concentrate = MLP(
                 [self.latent_dim] + self.decoder_hidden_layers + [self.input_size],
@@ -441,8 +449,14 @@ class SURE(nn.Module):
                 zns = pyro.sample('zn', dist.Gumbel(zn_loc, zn_scale).to_event(1))
 
             if self.cell_factor_size>0:
-                zus = self.cell_factor_effect(us)
-                zs = zns + zus
+                #zus = self.decoder_undesired([zns,us])
+                zus = None
+                for i in np.arange(self.cell_factor_size):
+                    if i==0:
+                        zus = self.cell_factor_effect[i]([zns,us[:,i].reshape(-1,1)])
+                    else:
+                        zus = zus + self.cell_factor_effect[i]([zns,us[:,i].reshape(-1,1)])
+                zs = zns+zus
             else:
                 zs = zns
 
@@ -631,8 +645,14 @@ class SURE(nn.Module):
                     zns = pyro.sample('zn', dist.Gumbel(zn_loc, zn_scale).to_event(1), obs=embeds)
 
             if self.cell_factor_size>0:
-                zus = self.cell_factor_effect(us)
-                zs = zus + zns
+                #zus = self.decoder_undesired([zns,us])
+                zus = None
+                for i in np.arange(self.cell_factor_size):
+                    if i==0:
+                        zus = self.cell_factor_effect[i]([zns,us[:,i].reshape(-1,1)])
+                    else:
+                        zus = zus + self.cell_factor_effect[i]([zns,us[:,i].reshape(-1,1)])
+                zs = zns+zus
             else:
                 zs = zns
 
@@ -793,6 +813,42 @@ class SURE(nn.Module):
         A = np.concatenate(A)
         return A
     
+    def _cell_move(self, xs, factor_idx, perturb):
+        zns = self.encoder_zn(xs)
+        if type(factor_idx) == str:
+            factor_idx = np.where(self.cell_factor_names==factor_idx)
+        
+        if perturb.ndim==2:
+            ms = self.cell_factor_effect[factor_idx]([zns, perturb])
+        else:
+            ms = self.cell_factor_effect[factor_idx]([zns, perturb.reshape(-1,1)])
+            
+        return ms 
+
+    def get_cell_move(self, 
+                             xs, 
+                             factor_idx,
+                             perturb,
+                             batch_size: int = 1024):
+        """
+        Return cells' changes in the latent space induced by specific perturbation of a factor
+
+        """
+        xs = self.preprocess(xs)
+        xs = convert_to_tensor(xs, device=self.get_device())
+        dataset = CustomDataset(xs)
+        dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=False)
+
+        Z = []
+        with tqdm(total=len(dataloader), desc='', unit='batch') as pbar:
+            for X_batch, _ in dataloader:
+                zns = self._cell_move(X_batch, factor_idx, perturb)
+                Z.append(tensor_to_numpy(zns))
+                pbar.update(1)
+
+        Z = np.concatenate(Z)
+        return Z
+    
     def preprocess(self, xs, threshold=0):
         if self.loss_func == 'bernoulli':
             ad = sc.AnnData(xs)
@@ -931,13 +987,6 @@ class SURE(nn.Module):
                 # Update progress bar
                 pbar.set_postfix({'loss': str_loss})
                 pbar.update(1)
-        
-        assigns = self.soft_assignments(xs)
-        assigns = convert_to_tensor(assigns, dtype=self.dtype, device=self.get_device())
-        self.codebook_weights = torch.sum(assigns, dim=0)
-        self.codebook_weights = self.codebook_weights / torch.sum(self.codebook_weights)
-        #self.inverse_dispersion = pyro.param('inverse_dispersion').item()
-        #self.dof = pyro.param('dof').item()
 
     @classmethod
     def save_model(cls, model, file_path, compression=False):
